#include <unistd.h>
#include <fcntl.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <sys/ioctl.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <linux/../scsi/sg.h>  /* cope with silly includes */
#include "sg_err.h"

/* Test code for D. Gilbert's extensions to the Linux OS SCSI generic ("sg")
   device driver.
*  Copyright (C) 1999 D. Gilbert
*  This program is free software; you can redistribute it and/or modify
*  it under the terms of the GNU General Public License as published by
*  the Free Software Foundation; either version 2, or (at your option)
*  any later version.

   This program scans the "sg" device space (ie actual + simulated SCSI
   generic devices). 
   Options: -w   open writable (new driver opens readable unless -i)
            -n   numeric scan: scan /dev/sg0,1,2, ....
            -i   do SCSI inquiry on device (implies -w)

   By default this program will look for /dev/sga first and if not found
   will turn on "-n" and try again.

   Note: This program is written to work under both the original and
   the new sg driver.

   Version 0.94 990519


6 byte INQUIRY command:
[0x12][   |lu][pg cde][res   ][al len][cntrl ]
6 byte commands [READ: 0x08, WRITE: 0x0a]:
[cmd ][had|lu][midAdd][lowAdd][count ][flags ]
10 byte commands [EREAD: 0x28, EWRITE: 0x2a, READ_CAPACITY 0x25]:
[cmd ][   |lu][hiAddr][hmAddr][lmAddr][lowAdd][      ][hiCnt ][lowCnt][flags ]
12 byte commands [LREAD: 0xd8, LWRITE: 0xda]:
[cmd ][   |lu][hiAddr][hmAddr][lmAddr][lowAdd][hiCnt ][hmCnt ][lmCnt ][lowCnt]
 ... [      ][flags ]
*/
        

#define OFF sizeof(struct sg_header)
#define INQ_REPLY_LEN 96        /* logic assumes >= sizeof(inqCmdBlk) */
#define INQ_CMD_LEN 6


#ifdef SG_SCATTER_SZ
#define OPEN_FLAG O_RDONLY
#else
#define OPEN_FLAG O_RDWR
#endif

typedef struct my_scsi_idlun { 
/* why can't userland see this structure ??? */
    int dev_id;
    int host_unique_id;
} My_scsi_idlun;

void usage()
{
    printf("Usage: 'sg_scan [-n] [-w]'\n");
    printf("    where: -n   to numeric scan (ie sg0, sg1, sg2)\n");
    printf("           -w   force open with read/write flag\n");
    printf("           -i   do SCSI inquiry, output results\n");
}

void make_dev_name(char * fname, int k, int do_numeric)
{
    char buff[64];

    strcpy(fname, "/dev/sg");
    if (do_numeric) {
        sprintf(buff, "%d", k);
        strcat(fname, buff);
    }
    else {
        if (k <= 26) {
            buff[0] = 'a' + (char)k;
            buff[1] = '\0';
            strcat(fname, buff);
        }
        else
            strcat(fname, "xxxx");
    }
}

#define EBUFF_LEN 256

int main(int argc, char * argv[])
{
    int sg_fd, res, k, f;
    unsigned char inqCmdBlk [INQ_CMD_LEN] = 
                                {0x12, 0, 0, 0, INQ_REPLY_LEN, 0};
    unsigned char * inqBuff = malloc(OFF + INQ_REPLY_LEN);
    int inqInLen = OFF + sizeof(inqCmdBlk);
    int inqOutLen = OFF + INQ_REPLY_LEN;
    unsigned char * buffp = inqBuff + OFF;
    struct sg_header * isghp = (struct sg_header *)inqBuff;
    int do_numeric = 0;
    int do_inquiry = 0;
    int writeable = 0;
    char fname[64];
    char ebuff[EBUFF_LEN];
    My_scsi_idlun my_idlun;
    int host_no;
    int flags;
    int emul;

    for (k = 1; k < argc; ++k) {
        if (0 == strcmp("-n", argv[k]))
            do_numeric = 1;
        else if (0 == strcmp("-w", argv[k]))
            writeable = 1;
        else if (0 == strcmp("-i", argv[k])) {
            do_inquiry = 1;
            writeable = 1;
        }
        else if (*argv[k] == '-') {
            printf("Unknown switch\n");
            usage();
            return 1;
        }
        else if (*argv[k] != '-') {
            printf("Unknown argument\n");
            usage();
            return 1;
        }
    }

    flags = writeable ? O_RDWR : OPEN_FLAG;
    
    while (1) {
        for (k = 0, res = 0; k < 1000; 
             ++k, res = (sg_fd >= 0) ? close(sg_fd) : 0) {
            if (res < 0) {
                sprintf(ebuff, "Error closing %s ", fname);
                perror("sg_scan: close error");
                return 1;
            }
            make_dev_name(fname, k, do_numeric);

            sg_fd = open(fname, flags | O_NONBLOCK);
            if (sg_fd < 0) {
                if (EBUSY == errno) {
                    printf("%s: device busy (O_EXCL lock), skipping\n", fname);
                    continue;
                }
                else if ((ENODEV == errno) || (ENOENT == errno) ||
                         (ENXIO == errno))
                    break;
                else {
                    sprintf(ebuff, "Error opening %s ", fname);
                    perror(ebuff);
                    continue;
                }
            }
            res = ioctl(sg_fd, SCSI_IOCTL_GET_IDLUN, &my_idlun);
            if (res < 0) {
                sprintf(ebuff, "device %s failed on scsi ioctl, skip", fname);
                perror(ebuff);
                continue;
            }
            res = ioctl(sg_fd, SCSI_IOCTL_GET_BUS_NUMBER, &host_no);
            if (res < 0) {
                sprintf(ebuff, "device %s failed on scsi ioctl(2), skip",
                        fname);
                perror(ebuff);
                continue;
            }
            res = ioctl(sg_fd, SG_EMULATED_HOST, &emul);
            if (res < 0) {
                sprintf(ebuff, "device %s failed on sg ioctl(3), skip", fname);
                perror(ebuff);
                continue;
            }
            printf("%s: scsi%d, channel=%d, id=%d, lun=%d", fname, host_no,
                   (my_idlun.dev_id >> 16) & 0xff, my_idlun.dev_id & 0xff, 
                   (my_idlun.dev_id >> 8) & 0xff);
            if (emul)
                printf(" [em]");
    #if 0
            printf(", huid=%d", my_idlun.host_unique_id);
    #endif
    #ifdef SG_SCATTER_SZ
            {
                Sg_scsi_id m_id;

                res = ioctl(sg_fd, SG_GET_SCSI_ID, &m_id);
                if (res < 0) {
                    sprintf(ebuff, "device %s ioctls(4), skip", fname);
                    perror(ebuff);
                    continue;
                }
                printf(", type=%d\n", m_id.scsi_type);
            }
    #else
            printf("\n");
    #endif
            if (! do_inquiry)
                continue;
            
            memset(isghp, 0, sizeof(struct sg_header));
            isghp->reply_len = inqOutLen;
            memcpy(inqBuff + OFF, inqCmdBlk, INQ_CMD_LEN);

            f = fcntl(sg_fd, F_GETFL);
            fcntl(sg_fd, F_SETFL, f & (~ O_NONBLOCK)); /* turn on blocking */

            res = write(sg_fd, inqBuff, inqInLen);
            if (res < 0) {
                sprintf(ebuff, "device %s writing, skip", fname);
                perror(ebuff);
                continue;
            }
            res = read(sg_fd, inqBuff, inqOutLen);
            if (res < 0) {
                sprintf(ebuff, "device %s reading, skip", fname);
                perror(ebuff);
                continue;
            }
    #ifdef SG_SCATTER_SZ
            if (! sg_chk_n_print("Error from Inquiry", isghp->target_status, 
                                 isghp->host_status, isghp->driver_status, 
                                 isghp->sense_buffer))
                continue;
    #else
            if ((isghp->result != 0) || (0 != isghp->sense_buffer[0])) {
                printf("Error from Inquiry: result=%d\n", isghp->result);
                if (0 != isghp->sense_buffer[0])
                    sg_print_sense("Error from Inquiry", isghp->sense_buffer);
                continue;
            }
    #endif
            f = (int)*(buffp + 7);
            printf("    %.8s  %.16s  %.4s, byte_7=0x%x\n", buffp + 8, 
                   buffp + 16, buffp + 32, f);
        }
        if ((0 == k) && (! do_numeric)) {
            printf("Didn't find /dev/sga so try numeric scan ...\n");
            do_numeric = 1;
        }
        else
           break;
    }
    return 0;
}
